"""

Helpers for tests.

"""

import ast
from typing import Any, Tuple, Callable, TypeVar
from ast_decompiler import decompile
from ast_decompiler.check import check as check
import difflib
import sys

VERSION = sys.version_info.major

_CallableT = TypeVar("_CallableT", bound=Callable[..., None])


def assert_decompiles(
    code: str,
    result: str,
    do_check: bool = True,
    indentation: int = 4,
    line_length: int = 100,
    starting_indentation: int = 0,
) -> None:
    """Asserts that code, when parsed, decompiles into result."""
    decompile_result = decompile(
        ast.parse(code),
        indentation=indentation,
        line_length=line_length,
        starting_indentation=starting_indentation,
    )
    if do_check:
        check(decompile_result)
    if result != decompile_result:
        print(">>> expected")
        print(result)
        print(">>> actual")
        print(decompile_result)
        print(">>> diff")
        for line in difflib.unified_diff(
            result.splitlines(), decompile_result.splitlines()
        ):
            print(line)
        assert False, f"failed to decompile {code}"


def only_on_version(py_version: int) -> Callable[[_CallableT], _CallableT]:
    """Decorator that runs a test only if the Python version matches."""
    if py_version != VERSION:

        def decorator(fn: Callable[..., Any]) -> Callable[..., None]:
            return lambda *args: None

    else:

        def decorator(fn: _CallableT) -> _CallableT:
            return fn

    return decorator


def skip_before(py_version: Tuple[int, int]) -> Callable[[_CallableT], _CallableT]:
    """Decorator that skips a test on Python versions before py_version."""
    if sys.version_info < py_version:

        def decorator(fn: Callable[..., Any]) -> Callable[..., None]:
            return lambda *args: None

    else:

        def decorator(fn: _CallableT) -> _CallableT:
            return fn

    return decorator


def skip_after(py_version: Tuple[int, int]) -> Callable[[_CallableT], _CallableT]:
    """Decorator that skips a test on Python versions after py_version."""
    if sys.version_info > py_version:

        def decorator(fn: Callable[..., Any]) -> Callable[..., None]:
            return lambda *args, **kwargs: None

    else:

        def decorator(fn: _CallableT) -> _CallableT:
            return fn

    return decorator



import ast
from ast_decompiler import decompile
import difflib


def check(code: str) -> None:
    """Checks that the code remains the same when decompiled and re-parsed."""
    tree = ast.parse(code)
    new_code = decompile(tree)
    try:
        new_tree = ast.parse(new_code)
    except SyntaxError as e:
        if e.lineno is None:
            raise
        print(">>> syntax error:")
        lineno = e.lineno - 1
        min_lineno = max(0, lineno - 3)
        max_lineno = lineno + 3
        for line in new_code.splitlines()[min_lineno:max_lineno]:
            print(line)
        raise

    dumped = ast.dump(ast.parse(code))
    new_dumped = ast.dump(new_tree)

    if dumped != new_dumped:
        print(code)
        print(new_code)
        for line in difflib.unified_diff(dumped.split(), new_dumped.split()):
            print(line)
        assert False, f"{dumped} != {new_dumped}"



from .tests import check, skip_before, skip_after


@skip_before((3, 5))
def test_MatMult() -> None:
    check("a @ b")
    check("(a * b) @ c")
    check("a * (b @ c)")
    check("a + (b @ c)")


@skip_before((3, 5))
def test_AsyncFunctionDef() -> None:
    check(
        """
async def f(a, b):
    pass
"""
    )


@skip_before((3, 6))
def test_async_gen() -> None:
    check(
        """
async def f():
    yield
"""
    )


@skip_before((3, 6))
def test_async_comprehensions() -> None:
    check(
        """
async def f(lst):
    return [await x for x in lst]
"""
    )
    check(
        """
async def f(lst):
    a = [x async for x in lst]
    b = {x async for x in lst}
    c = {x: x async for x in lst}
    d = (x async for x in lst)
"""
    )


@skip_before((3, 0))
def test_annotations() -> None:
    # TODO test precedence
    check(
        """
def f(a: int, b: str) -> float:
    pass
"""
    )


@skip_before((3, 0))
def test_class_keywords() -> None:
    check(
        """
class Foo(a=3):
    pass
"""
    )
    check(
        """
class WithMeta(metaclass=type):
    pass
"""
    )


@skip_before((3, 0))
def test_annotations() -> None:
    check("def f(a: int, b: int = 0, *args: int, c: int, **kwargs: int): pass")


@skip_before((3, 5))
def test_AsyncFor() -> None:
    check(
        """
async def f(y):
    async for x in y:
        pass
"""
    )


@skip_before((3, 0))
def test_py3_with() -> None:
    check(
        """
with a as b:
    pass
"""
    )
    check(
        """
with a as b, c as d:
    pass
"""
    )
    check(
        """
with a as b:
    with c as d:
        pass
"""
    )


@skip_before((3, 5))
def test_async_with() -> None:
    check(
        """
async def f(a):
    async with a as b:
        pass
"""
    )


@skip_before((3, 0))
def test_raise_with_cause() -> None:
    check(
        """
raise e from ee
"""
    )


@skip_before((3, 0))
def test_Nonlocal() -> None:
    check(
        """
def f(x):
    nonlocal y
"""
    )
    check(
        """
def f(x):
    nonlocal y, z
"""
    )


@skip_before((3, 5))
def test_Await() -> None:
    check(
        """
async def f(x):
    await x
"""
    )
    check(
        """
async def f(x):
    1 + await x
"""
    )
    check(
        """
async def f(x):
    x = await x
"""
    )
    check(
        """
async def f(x):
    x += await x
"""
    )
    check(
        """
async def f(x):
    return 3, (await x)
"""
    )


@skip_before((3, 0))
def test_YieldFrom() -> None:
    check("yield from x")
    check("1 + (yield from x)")
    check("x = yield from x")
    check("x += yield from x")
    check("return 3, (yield from x)")


@skip_before((3, 6))
def test_FormattedValue() -> None:
    check('f"a"')
    check('f"{b}"')
    check('f"{b}a"')
    check('f"{b!r}"')
    check('f"{b:a}"')
    check('f"{b!r:a}"')
    check("f\"{'b'!r:'a'}\"")
    check('f"{a}b{c!a}s"')
    check('f"{a.b}c{d()}"')
    check("f'{1/3:.1f}'")
    check(r"f'{a}\''")
    check("f'{1/3:{5}.1}'")
    check("f'{ {a, b, c} }'")
    check("f'{ {a: b} }'")
    check("f'{ {a for a in b} }'")
    check("f'{ {a: b for a, b in c} }'")
    check(r"f'{a}\n'")
    check(r"f'{a}\t'")
    check("f'{a}é'")
    check('f"{{"')
    check('f"}}"')
    check('f"{{{a}"')


@skip_before((3, 0))
def test_Bytes() -> None:
    check('b"a"')


@skip_before((3, 0))
def test_NameConstant() -> None:
    check("True")


@skip_before((3, 0))
def test_Starred() -> None:
    check("a, *b = 3")
    check("[a, *b]")


@skip_before((3, 0))
def test_kwonlyargs() -> None:
    check("def f(a, *, b): pass")
    check("def f(a, *args, b): pass")
    check("def f(a, *, b=3): pass")
    check("def f(a, *args, b=3): pass")
    check("def f(a, *args, b=3, **kwargs): pass")


@skip_before((3, 6))
def test_annassign() -> None:
    check("a: int")
    check("a: int = 3")
    check("(a): int")
    check(
        """
class A:
    b: int
"""
    )
    check(
        """
def f():
    a: int
"""
    )


@skip_before((3, 7))
def test_future_annotations() -> None:
    # This doesn't really affect ast_decompiler because the __future__
    # import doesn't change the AST.
    check(
        """
from __future__ import annotations

def f(x: int) -> str:
    pass

y: float
"""
    )


@skip_after((3, 6))
def test_async_varname() -> None:
    check("import async")
    check("await = 3")
    check(
        """
def async(async, await=3):
    return async + await
"""
    )


@skip_before((3, 7))
def test_async_await_in_fstring() -> None:
    check("f'{await x}'")
    check("f'{[x async for x in y]}'")


@skip_before((3, 7))
def test_too_many_args() -> None:
    args = ", ".join(f"x{i}" for i in range(300))
    check(
        """
def f({}):
    pass

f({})
""".format(
            args, args
        )
    )


def test_finally_continue() -> None:
    check(
        """
def f():
    for x in y:
        try:
            whatever
        finally:
            continue
"""
    )


@skip_before((3, 4))
def test_unpacking() -> None:
    check(
        """
def parse(family):
    lastname, *members = family.split()
    return (lastname.upper(), *members)
"""
    )


@skip_before((3, 8))
def test_unparenthesized_unpacking() -> None:
    check(
        """
def parse(family):
    lastname, *members = family.split()
    return lastname.upper(), *members
"""
    )


@skip_before((3, 8))
def test_assignment_expression() -> None:
    # Some of these can be used unparenthesized in 3.10+ but we don't bother.
    check(
        """
if (x := 3):
    pass
{(y := 4)}
{(z := 5) for a in b}
lst[(alpha := 3)]
lst[(beta := 4):(gamma := 5)]
"""
    )


@skip_before((3, 8))
def test_positional_only() -> None:
    check(
        """
def f(x, /):
    pass
"""
    )


@skip_before((3, 8))
def test_fstring_debug_specifier() -> None:
    check("f'{user=} {member_since=}'")
    check("f'{user=!s}  {delta.days=:,d}'")



from .tests import assert_decompiles


def check_split(original: str, multiline: str, length_reduction: int = 2) -> None:
    assert_decompiles(original, original, line_length=len(original))
    assert_decompiles(
        original, multiline, line_length=len(original.strip()) - length_reduction
    )


def test_with_prefix() -> None:
    prefixes = ["from x import"]
    for prefix in prefixes:
        check_split(
            f"{prefix} a, b, c\n",
            f"""{prefix} (
    a,
    b,
    c,
)
""",
        )


def test_del() -> None:
    original = "del a, b, c\n"
    check_split(original, original, length_reduction=10)


def test_import() -> None:
    original = "import a, b, c, d\n"
    check_split(original, original, length_reduction=10)


def test_global() -> None:
    original = "global a, b, c, d\n"
    check_split(original, original, length_reduction=10)


def test_boolop() -> None:
    check_split(
        "if a and b and c:\n    pass\n",
        """if (
    a and
    b and
    c
):
    pass
""",
        length_reduction=12,
    )


def test_display() -> None:
    delimiters = [("{", "}"), ("[", "]"), ("\n\nclass Foo(", "):\n    pass")]
    for start, end in delimiters:
        original = f"{start}a, b, c{end}\n"
        assert_decompiles(original, original, line_length=len(original))

        assert_decompiles(
            original,
            f"""{start}
    a,
    b,
    c,
{end}
""",
            line_length=len(start.lstrip()) + 5,
        )


def test_assign() -> None:
    check_split(
        "a, b, c = lst\n",
        """(
    a,
    b,
    c,
) = lst
""",
        length_reduction=7,
    )

    original = "a = b = c = 3\n"
    check_split(original, original, length_reduction=3)


def test_tuple() -> None:
    check_split(
        "a, b, c\n",
        """(
    a,
    b,
    c,
)
""",
    )


def test_extslice() -> None:
    check_split(
        "d[a:, b, c]\n",
        """d[
    a:,
    b,
    c,
]
""",
    )


def test_comprehension() -> None:
    check_split(
        "[x for y in lst for x in y]\n",
        """[
    x
    for y in lst
    for x in y
]
""",
    )


def test_dict() -> None:
    check_split(
        "{a: b, c: d}\n",
        """{
    a: b,
    c: d,
}
""",
    )


def test_dictcomp() -> None:
    check_split(
        "{a: b for a, b in c}\n",
        """{
    a: b
    for a, b in c
}
""",
    )


def test_function_def() -> None:
    check_split(
        "\ndef f(a, b, *args, **kwargs):\n    pass\n",
        """
def f(
    a,
    b,
    *args,
    **kwargs
):
    pass
""",
        length_reduction=12,
    )


def test_call() -> None:
    check_split(
        "f(a, b, **c)\n",
        """f(
    a,
    b,
    **c
)
""",
    )


def test_nesting() -> None:
    check_split(
        "f(f(a, b, c), g(d, e, f))\n",
        """f(
    f(a, b, c),
    g(d, e, f)
)
""",
        length_reduction=9,
    )



from .tests import assert_decompiles


def test_indentation() -> None:
    assert_decompiles(
        """
if x:
    pass
""",
        """if x:
 pass
""",
        indentation=1,
    )


def test_starting_indentation() -> None:
    assert_decompiles(
        """3""",
        """    3
""",
        starting_indentation=4,
        do_check=False,
    )



from .tests import assert_decompiles, only_on_version


def test_With() -> None:
    assert_decompiles(
        "with x as y, a as b: pass",
        """with x as y, a as b:
    pass
""",
    )


@only_on_version(2)
def test_With_collapsed() -> None:
    assert_decompiles(
        """
with x as y:
    with a as b:
        pass
""",
        """with x as y, a as b:
    pass
""",
    )


def test_TryFinally() -> None:
    assert_decompiles(
        """
try:
    1 / 0
except Exception as e:
    pass
else:
    z = 3
finally:
    z = 4
""",
        """try:
    1 / 0
except Exception as e:
    pass
else:
    z = 3
finally:
    z = 4
""",
    )


@only_on_version(2)
def test_TryFinally_collapsed() -> None:
    assert_decompiles(
        """
try:
    try:
        1 / 0
    except Exception as e:
        pass
finally:
    z = 4
""",
        """try:
    1 / 0
except Exception as e:
    pass
finally:
    z = 4
""",
    )


def test_If() -> None:
    assert_decompiles(
        """
if x: pass
else:
    if y: pass
    else: pass
""",
        """if x:
    pass
elif y:
    pass
else:
    pass
""",
    )
    assert_decompiles(
        """
if x: pass
elif a:
    if y: pass
    else:
        if z: pass
        else: pass
else:
    pass
""",
        """if x:
    pass
elif a:
    if y:
        pass
    elif z:
        pass
    else:
        pass
else:
    pass
""",
    )


def test_BinOp() -> None:
    assert_decompiles(
        """
f(a * b)
""",
        """f(a * b)
""",
    )
